import os
from yggdrasil.tests import extra_example
from yggdrasil.examples.tests import ExampleTstBase


@extra_example
class TestExampleRpcFib(ExampleTstBase):
    r"""Test the rpcFib example."""

    example_name = 'rpcFib'
    env = {'FIB_ITERATIONS': '3',
           'FIB_SERVER_SLEEP_SECONDS': '0.01'}
    
    @property
    def results(self):
        r"""Result that should be found in output files."""
        prev1 = 0
        prev2 = 1
        res = ''
        for i in range(int(self.env['FIB_ITERATIONS'])):
            next = prev1 + prev2
            res += 'fib(%2d<-) = %-2d<-\n' % ((i + 1), next)
            prev2 = prev1
            prev1 = next
        return [res]

    @property
    def output_files(self):
        r"""Output file."""
        return [os.path.join(self.tempdir, 'fibCli.txt')]



import numpy as np
from yggdrasil.tests import YggTestBase
from yggdrasil import units


class TestUnits(YggTestBase):
    r"""Tests for using pint for units."""

    def setup(self, *args, **kwargs):
        r"""Setup, create variables for testing."""
        self._vars_nounits = [1.0, np.zeros(5), int(1)]
        self._vars_units = [units.add_units(v, 'cm') for v in self._vars_nounits]
        super(TestUnits, self).setup(*args, **kwargs)

    def test_has_units(self):
        r"""Test has_units."""
        for v in self._vars_nounits:  # + ['string']:
            assert(not units.has_units(v))
        for v in self._vars_units:
            assert(units.has_units(v))

    def test_get_data(self):
        r"""Test get_data."""
        for v in self._vars_nounits:
            self.assert_equal(units.get_data(v), v)
        for vno, v in zip(self._vars_nounits, self._vars_units):
            self.assert_equal(units.get_data(v), np.array(vno))

    def test_get_units(self):
        r"""Test get_units."""
        for v in self._vars_nounits:
            self.assert_equal(units.get_units(v), '')
        for v in self._vars_units:
            self.assert_equal(units.get_units(v), str(units.as_unit('cm').units))

    def test_add_units(self):
        r"""Test add_units."""
        for v in self._vars_nounits:
            x = units.add_units(v, 'cm')
            assert(units.has_units(x))
        self.assert_equal(units.add_units(1.0, ''), 1.0)
        self.assert_equal(units.add_units(1.0, 'n/a'), 1.0)

    def test_is_null_unit(self):
        r"""Test is_null_unit."""
        assert(units.is_null_unit('n/a'))
        assert(units.is_null_unit(''))
        assert(not units.is_null_unit('cm'))

    def test_as_unit(self):
        r"""Test as_unit."""
        units.as_unit('cm')
        self.assert_raises(ValueError, units.as_unit, 'invalid')

    def test_is_unit(self):
        r"""Test is_unit."""
        assert(units.is_unit('n/a'))
        assert(units.is_unit(''))
        assert(units.is_unit('cm/s**2'))
        # Not supported by unyt
        # assert(units.is_unit('cm/s^2'))
        assert(units.is_unit('umol'))
        assert(units.is_unit('mmol'))
        assert(not units.is_unit('invalid'))

    def test_convert_to(self):
        r"""Test convert_to."""
        units.convert_to(1, 'm')
        for v in self._vars_units:
            units.convert_to(v, 'm')
            self.assert_raises(ValueError, units.convert_to, v, 's')

    def test_are_compatible(self):
        r"""Test are_compatible."""
        assert(units.are_compatible('cm', 'm'))
        assert(units.are_compatible('cm', ''))
        assert(not units.are_compatible('cm', 's'))
        assert(not units.are_compatible('cm', 'invalid'))
        assert(units.are_compatible('d', 'hr'))
        assert(units.are_compatible('hr', 'd'))



import yggdrasil.drivers.tests.test_CModelDriver as parent


class TestCPPModelParam(parent.TestCModelParam):
    r"""Test parameters for CPPModelDriver."""

    driver = 'CPPModelDriver'

    def __init__(self, *args, **kwargs):
        super(TestCPPModelParam, self).__init__(*args, **kwargs)
        self._inst_kwargs.setdefault('compiler_flags', [])
        self._inst_kwargs['compiler_flags'].append('-std=c++11')
    
    
class TestCPPModelDriverNoInit(TestCPPModelParam,
                               parent.TestCModelDriverNoInit):
    r"""Test runner for CPPModelDriver without init."""
    
    def test_write_try_except(self, **kwargs):
        r"""Test writing a try/except block."""
        kwargs.setdefault('error_type', '...')
        super(TestCPPModelDriverNoInit, self).test_write_try_except(**kwargs)
        

class TestCPPModelDriverNoStart(TestCPPModelParam,
                                parent.TestCModelDriverNoStart):
    r"""Test runner for CPPModelDriver without start."""
    pass
    
        
class TestCPPModelDriver(TestCPPModelParam, parent.TestCModelDriver):
    r"""Test runner for CPPModelDriver."""
    pass



import unittest
from yggdrasil import backwards
from yggdrasil.languages import install_languages


class TestInstallLanguages(unittest.TestCase):

    def test_update_argparser(self):
        r"""Test update_argparser."""
        install_languages.update_argparser(arglist=['all'])
        install_languages.update_argparser(arglist=['python'])
        install_languages.update_argparser(language='python',
                                           arglist=['python'])
        if not backwards.PY2:  # pragma: Python 3
            with self.assertRaises(SystemExit):
                install_languages.update_argparser(arglist=['-h'])
                unittest.main(exit=False)

    def test_install_language(self):
        r"""Test install_language."""
        install_languages.install_language('PYTHON', arglist=[])

    def test_install_all_languages(self):
        r"""Test install_all_languages."""
        install_languages.install_all_languages()
        install_languages.install_all_languages(from_setup=True)



from yggdrasil import backwards
from yggdrasil.serialize.DefaultSerialize import DefaultSerialize


class PickleSerialize(DefaultSerialize):
    r"""Class for serializing a python object into a bytes message by pickling.
    """

    _seritype = 'pickle'
    _schema_subtype_description = ('Serialize any Python object using a Python '
                                   'pickle.')
    _default_type = {'type': 'bytes'}
    is_framed = True

    def func_serialize(self, args):
        r"""Serialize a message.

        Args:
            args (obj): Python object to be serialized.

        Returns:
            bytes, str: Serialized message.

        """
        out = backwards.pickle.dumps(args)
        return backwards.as_bytes(out)

    def func_deserialize(self, msg):
        r"""Deserialize a message.

        Args:
            msg (str, bytes): Message to be deserialized.

        Returns:
            obj: Deserialized Python object.

        """
        out = backwards.pickle.loads(msg)
        return out

    @classmethod
    def get_first_frame(cls, msg):
        r"""Extract one frame from the provided message that may contain one
        or more frames.

        Args:
            msg (bytes): Message containing one or more frames.

        Returns:
            bytes: Portion of message containing the first frame. If no frames
                are found, an empty string will be returned.

        """
        fd = backwards.BytesIO(msg)
        try:
            backwards.pickle.load(fd)
            used = fd.tell()
        except BaseException:
            used = 0
        fd.close()
        return msg[:used]

    @classmethod
    def concatenate(cls, objects, **kwargs):
        r"""Concatenate objects to get object that would be recieved if
        the concatenated serialization were deserialized.

        Args:
            objects (list): Objects to be concatenated.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            list: Set of objects that results from concatenating those provided.

        """
        return objects
    
    @classmethod
    def get_testing_options(cls, **kwargs):
        r"""Method to return a dictionary of testing options for this class.

        Returns:
            dict: Dictionary of variables to use for testing.

        """
        out = super(PickleSerialize, cls).get_testing_options()
        if backwards.PY2:  # pragma: Python 2
            out['contents'] = ("S'Test message\\n'\np1\n."
                               + "S'Test message 2\\n'\np1\n.")
        else:  # pragma: Python 3
            out['contents'] = (b'\x80\x03C\rTest message\nq\x00.'
                               + b'\x80\x03C\x0fTest message 2\nq\x00.')
        return out



import copy
from yggdrasil.tests import YggTestClassInfo, assert_equal
from yggdrasil import tools
from yggdrasil.components import import_component
from yggdrasil.serialize import SerializeBase
from yggdrasil.metaschema.datatypes import encode_type


def test_demote_string():
    r"""Test format str creation of typedef."""
    x = SerializeBase.SerializeBase(format_str='%s')
    assert_equal(x.typedef, {'type': 'array',
                             'items': [{'type': 'bytes'}]})


class TestSerializeBase(YggTestClassInfo):
    r"""Test class for SerializeBase class."""

    _cls = 'SerializeBase'
    _empty_msg = b''
    attr_list = (copy.deepcopy(YggTestClassInfo.attr_list)
                 + ['datatype', 'typedef', 'numpy_dtype'])

    def __init__(self, *args, **kwargs):
        super(TestSerializeBase, self).__init__(*args, **kwargs)
        self._header_info = dict(arg1='1', arg2='two')

    @property
    def mod(self):
        r"""Module for class to be tested."""
        return 'yggdrasil.serialize.%s' % self.cls

    @property
    def inst_kwargs(self):
        r"""Keyword arguments for creating the test instance."""
        out = super(TestSerializeBase, self).inst_kwargs
        out.update(self.testing_options['kwargs'])
        return out

    def empty_head(self, msg):
        r"""dict: Empty header for message only contains the size."""
        out = dict(size=len(msg), incomplete=False)
        if msg == tools.YGG_MSG_EOF:  # pragma: debug
            out['eof'] = True
        return out

    def map_sent2recv(self, obj):
        r"""Convert a sent object into a received one."""
        return obj

    def assert_result_equal(self, x, y):
        r"""Assert that serialized/deserialized objects equal."""
        self.assert_equal(x, self.map_sent2recv(y))

    def test_field_specs(self):
        r"""Test field specifiers."""
        self.assert_equal(self.instance.numpy_dtype,
                          self.testing_options['dtype'])
        self.assert_equal(self.instance.extra_kwargs,
                          self.testing_options['extra_kwargs'])
        self.assert_equal(self.instance.typedef,
                          self.testing_options['typedef'])
        if isinstance(self.instance.typedef.get('items', []), dict):
            self.assert_raises(Exception, self.instance.get_field_names)
            self.assert_raises(Exception, self.instance.get_field_units)
        else:
            self.assert_equal(self.instance.get_field_names(),
                              self.testing_options.get('field_names', None))
            self.assert_equal(self.instance.get_field_units(),
                              self.testing_options.get('field_units', None))

    def test_concatenation(self):
        r"""Test message concatenation."""
        for x, y in self.testing_options.get('concatenate', []):
            self.assert_equal(self.instance.concatenate(x), y)
        
    def test_serialize(self):
        r"""Test serialize/deserialize."""
        for iobj in self.testing_options['objects']:
            if (self._cls == 'SerializeBase'):
                self.assert_raises(NotImplementedError, self.instance.serialize, iobj)
                self.assert_raises(NotImplementedError, self.instance.deserialize,
                                   b'test')
                break
            msg = self.instance.serialize(iobj)
            iout, ihead = self.instance.deserialize(msg)
            self.assert_result_equal(iout, iobj)
            # self.assert_equal(ihead, self.empty_head(msg))

    def test_serialize_no_metadata(self):
        r"""Test serializing without metadata."""
        if (self._cls == 'SerializeBase'):
            return
        self.instance.serialize(self.testing_options['objects'][0],
                                no_metadata=True)

    def test_deserialize_error(self):
        r"""Test error when deserializing message that is not bytes."""
        self.assert_raises(TypeError, self.instance.deserialize, None)
        
    def test_serialize_sinfo(self):
        r"""Test serialize/deserialize with serializer info."""
        if (self._cls == 'SerializeBase'):
            return
        hout = copy.deepcopy(self._header_info)
        hout.update(self.instance.serializer_info)
        temp_seri = import_component(
            'serializer', self.instance.serializer_info['seritype'])()
        for iobj in self.testing_options['objects']:
            hout.update(encode_type(iobj, typedef=self.instance.typedef))
            msg = self.instance.serialize(iobj, header_kwargs=self._header_info,
                                          add_serializer_info=True)
            iout, ihead = self.instance.deserialize(msg)
            hout.update(size=ihead['size'], id=ihead['id'],
                        incomplete=False)
            self.assert_result_equal(iout, iobj)
            self.assert_equal(ihead, hout)
            # Use info to reconstruct serializer
            iout, ihead = temp_seri.deserialize(msg)
            self.assert_result_equal(iout, iobj)
            self.assert_equal(ihead, hout)
            new_seri = import_component('serializer',
                                        ihead.pop('seritype', None))(**ihead)
            iout, ihead = new_seri.deserialize(msg)
            self.assert_result_equal(iout, iobj)
            self.assert_equal(ihead, hout)
            
    def test_serialize_header(self):
        r"""Test serialize/deserialize with header."""
        if (self._cls == 'SerializeBase'):
            return
        for iobj in self.testing_options['objects']:
            msg = self.instance.serialize(iobj, header_kwargs=self._header_info)
            iout, ihead = self.instance.deserialize(msg)
            self.assert_result_equal(iout, iobj)
            # self.assert_equal(ihead, self._header_info)
        
    def test_serialize_eof(self):
        r"""Test serialize/deserialize EOF."""
        if (self._cls == 'SerializeBase'):
            return
        iobj = tools.YGG_MSG_EOF
        msg = self.instance.serialize(iobj)
        iout, ihead = self.instance.deserialize(msg)
        self.assert_equal(iout, iobj)
        # self.assert_equal(ihead, self.empty_head(msg))
        
    def test_serialize_eof_header(self):
        r"""Test serialize/deserialize EOF with header."""
        if (self._cls == 'SerializeBase'):
            return
        iobj = tools.YGG_MSG_EOF
        msg = self.instance.serialize(iobj, header_kwargs=self._header_info)
        iout, ihead = self.instance.deserialize(msg)
        self.assert_equal(iout, iobj)
        # self.assert_equal(ihead, self.empty_head(msg))
        
    def test_deserialize_empty(self):
        r"""Test call for empty string."""
        if (self._cls == 'SerializeBase'):
            return
        out, head = self.instance.deserialize(self._empty_msg)
        self.assert_result_equal(out, self.testing_options['empty'])
        self.assert_equal(head, self.empty_head(self._empty_msg))



import copy
import numpy as np
from yggdrasil.tests import assert_raises
from yggdrasil.metaschema.datatypes.ContainerMetaschemaType import (
    ContainerMetaschemaType)


_vallist = [np.float32(1),
            b'hello', u'hello',
            {'nested': np.int64(2)},
            [np.complex128(4), np.uint8(0)]]
_deflist = [{'type': 'float',
             'precision': 32,
             'units': ''},
            {'type': 'bytes',
             'precision': 40,
             'units': ''},
            {'type': 'unicode',
             'precision': 40,
             'units': ''},
            {'type': 'object',
             'properties': {'nested': {'type': 'int',
                                       'precision': 64,
                                       'units': ''}}},
            {'type': 'array',
             'items': [{'type': 'complex',
                        'precision': 128,
                        'units': ''},
                       {'type': 'uint',
                        'precision': 8,
                        'units': ''}]}]
_typedef = []
for v in _deflist:
    itypedef = {'type': v['type']}
    if v['type'] == 'object':
        itypedef['properties'] = copy.deepcopy(v['properties'])
    elif v['type'] == 'array':
        itypedef['items'] = copy.deepcopy(v['items'])
    _typedef.append(itypedef)
_count = len(_vallist)


def test_container_errors():
    r"""Test implementation errors on bare container class."""
    assert_raises(NotImplementedError, ContainerMetaschemaType._iterate, None)
    assert_raises(NotImplementedError, ContainerMetaschemaType._assign,
                  None, None, None)
    assert_raises(NotImplementedError, ContainerMetaschemaType._has_element,
                  None, None)



import yggdrasil.drivers.tests.test_InterpretedModelDriver as parent


class TestPythonModelParam(parent.TestInterpretedModelParam):
    r"""Test parameters for PythonModelDriver."""

    driver = "PythonModelDriver"


class TestPythonModelDriverNoInit(TestPythonModelParam,
                                  parent.TestInterpretedModelDriverNoInit):
    r"""Test runner for PythonModelDriver without init."""
    
    def test_write_model_wrapper(self):
        r"""Test writing a model based on yaml parameters."""
        inputs = [{'name': 'a', 'type': 'bytes', 'outside_loop': True},
                  {'name': 'b', 'type': {'type': 'int', 'precision': 64}}]
        outputs = [{'name': 'y', 'type': {'type': 'float', 'precision': 32}},
                   {'name': 'z', 'type': 'bytes', 'outside_loop': True}]
        self.import_cls.write_model_wrapper(None, 'test',
                                            inputs=inputs,
                                            outputs=outputs)
        super(TestPythonModelDriverNoInit, self).test_write_model_wrapper()
        

class TestPythonModelDriverNoStart(TestPythonModelParam,
                                   parent.TestInterpretedModelDriverNoStart):
    r"""Test runner for PythonModelDriver without start."""
    pass


class TestPythonModelDriver(TestPythonModelParam,
                            parent.TestInterpretedModelDriver):
    r"""Test runner for PythonModelDriver."""
    pass



r"""This module allows for backward compatibility."""
import sys
import time
import base64
from collections import OrderedDict
from yggdrasil.scanf import scanf
_python_version = '%d.%d' % (sys.version_info[0], sys.version_info[1])
PY2 = (sys.version_info[0] == 2)
PY34 = ((sys.version_info[0] == 3) and (sys.version_info[1] == 4))
if PY2:  # pragma: Python 2
    import cPickle as pickle
    import ConfigParser as configparser
    import StringIO as sio
    from __builtin__ import unicode
    import types
    BytesIO = sio.StringIO
    StringIO = sio.StringIO
    file_type = types.FileType
    bytes_type = str
    unicode_type = unicode
    string_type = str
    np_dtype_str = 'S'
    string_types = (str, unicode, bytearray)
    class_types = (type, types.ClassType)
    base64_encode = base64.encodestring
    base64_decode = base64.decodestring
else:  # pragma: Python 3
    import pickle
    import configparser
    import io as sio
    BytesIO = sio.BytesIO
    StringIO = sio.StringIO
    file_type = sio.IOBase
    bytes_type = bytes
    unicode_type = str
    string_type = str
    unicode = None
    np_dtype_str = 'S'
    string_types = (bytes, str, bytearray)
    class_types = (type, )
    base64_encode = base64.encodebytes
    base64_decode = base64.decodebytes
if sys.version_info >= (3, 3):
    clock_time = time.perf_counter
else:
    clock_time = time.clock


def scanf_bytes(fmt, bytes_line):
    r"""Extract parameters from a bytes object using scanf."""
    if PY2:  # pragma: Python 2
        out_byt = scanf(fmt, bytes_line)
    else:  # pragma: Python 3
        out_str = scanf(as_str(fmt), as_str(bytes_line))
        out_byt = match_stype(bytes_line, out_str, recurse=True,
                              allow_pass=True)
    return out_byt


def assert_str(s):
    r"""Assert that the input is in str type appropriate for the version of
    Python.

    Arguments:
        s (obj): Object to be tested if it is of the proper str class.

    Raises:
        AssertionError: If the object is not in the proper str class.

    """
    assert(isinstance(s, string_type))


def assert_bytes(s):
    r"""Assert that the input is in bytes appropriate for the version of
    Python.

    Arguments:
        s (obj): Object to be tested if it is of the proper bytes class.

    Raises:
        AssertionError: If the object is not in the proper bytes class.

    """
    assert(isinstance(s, bytes_type))


def assert_unicode(s):
    r"""Assert that the input is in unicode appropriate for the version of
    Python.

    Arguments:
        s (obj): Object to be tested if it is of the proper unicode class.

    Raises:
        AssertionError: If the object is not in the proper unicode class.

    """
    assert(isinstance(s, unicode_type))


def recurse_conv(s, func_conv, **kwargs):
    r"""Recurse into list, tuples, or dicts, applying a conversion funciton.

    Args:
        s (object): Object to apply conversion to.
        func_conv (function): Callable conversion function that takes the object
            as input and returns a converted version of it.
        **kwargs: Additional keyword arguments are passed to the conversion
            function.

    Returns:
        object: Converted version of the object.

    Raises:
        TypeError: If the object is not a list, tuple, or dictionary.

    """
    if isinstance(s, list):
        s_out = []
        for i in range(len(s)):
            s_out.append(func_conv(s[i], **kwargs))
    elif isinstance(s, tuple):
        s_out = tuple(func_conv(list(s), **kwargs))
    elif isinstance(s, (dict, OrderedDict)):
        s_out = {}
        for k0 in s.keys():
            k = func_conv(k0, **kwargs)
            s_out[k] = func_conv(s[k0], **kwargs)
    else:
        raise TypeError("Cannot recurse into type: %s" % type(s))
    return s_out


def as_unicode(s_in, recurse=False, convert_types=None, allow_pass=False):
    r"""Convert from bytes/unicode/str to unicode.

    Arguments:
        s_in (bytes, unicode, str): Object to be converted into unicode.
        recurse (bool, optional): If True, objects that are lists, tuples, or
            dicts will be recursed into, converting string elements into
            unicode.
        convert_types (list, optional): Python types that should be converted
            to unicode. Defaults to the string types for the current Python
            version ([str, unicode, bytearray] in Python 2, [bytes, str,
            bytearray] otherwise).
        allow_pass (bool, optional): If True and the provided object is not
            converted, no error will be raised. Defaults to False and objects
            that cannot be converted will raise an error.

    Returns:
        unicode/str: Unicode version of input (unicode in Python 2, str in
            Python 3).

    Raises:
       TypeError: If supplied type cannot be converted to unicode and allow_pass
           is False.

    """
    if convert_types is None:
        convert_types = string_types
    if isinstance(s_in, convert_types):
        if PY2:  # pragma: Python 2
            if isinstance(s_in, (str, bytearray)):
                s_out = unicode_type(s_in)
            elif isinstance(s_in, unicode):
                s_out = s_in
            else:
                s_out = unicode_type(s_in)
        else:  # pragma: Python 3
            if isinstance(s_in, str):
                s_out = s_in
            elif isinstance(s_in, (bytes, bytearray)):
                s_out = s_in.decode("utf-8")
            else:
                s_out = unicode_type(s_in)
    elif recurse and isinstance(s_in, (list, tuple, dict, OrderedDict)):
        s_out = recurse_conv(s_in, as_unicode, recurse=True,
                             convert_types=convert_types, allow_pass=allow_pass)
    elif allow_pass:
        s_out = s_in
    else:
        raise TypeError("Cannot convert type %s to unicode." % type(s_in)
                        + "Must be one of: %s" % str(convert_types))
    return s_out


def as_bytes(s_in, recurse=False, convert_types=None, allow_pass=False):
    r"""Convert from bytes/unicode/str to a bytes object.

    Arguments:
        s_in (str, bytes, unicode): Object to convert to a bytes version.
        recurse (bool, optional): If True, objects that are lists, tuples, or
            dicts will be recursed into, converting string elements into
            unicode.
        convert_types (list, optional): Python types that should be converted
            to unicode. Defaults to the string types for the current Python
            version ([str, unicode, bytearray] in Python 2, [bytes, str,
            bytearray] otherwise).
        allow_pass (bool, optional): If True and the provided object is not
            converted, no error will be raised. Defaults to False and objects
            that cannot be converted will raise an error.

    Returns:
        bytes/str: Bytes version of input (str in Python 2, bytes in Python 3).

    Raises:
       TypeError: If supplied type cannot be converted to bytes.

    """
    if convert_types is None:
        convert_types = string_types
    if isinstance(s_in, convert_types):
        if PY2:  # pragma: Python 2
            if isinstance(s_in, bytearray):
                s_out = bytes_type(s_in)  # In python 2 str is bytes
            elif isinstance(s_in, str):
                s_out = s_in
            elif isinstance(s_in, unicode):
                s_out = s_in.encode("utf-8")
            else:
                s_out = str(s_in)
        else:  # pragma: Python 3
            if isinstance(s_in, bytes):
                s_out = s_in
            elif isinstance(s_in, bytearray):
                s_out = bytes_type(s_in)
            elif isinstance(s_in, str):
                s_out = s_in.encode("utf-8")
            else:
                s_out = as_bytes(str(s_in))
    elif recurse and isinstance(s_in, (list, tuple, dict, OrderedDict)):
        s_out = recurse_conv(s_in, as_bytes, recurse=True,
                             convert_types=convert_types, allow_pass=allow_pass)
    elif allow_pass:
        s_out = s_in
    else:
        raise TypeError("Cannot convert type %s to bytes." % type(s_in)
                        + "Must be one of: %s" % str(convert_types))
    return s_out
    

def as_str(s_in, recurse=False, convert_types=None, allow_pass=False):
    r"""Convert from bytes/unicode/str to a str object.

    Arguments:
        s_in (str, bytes, unicode): Object to convert to a str version.
        recurse (bool, optional): If True, objects that are lists, tuples, or
            dicts will be recursed into, converting string elements into
            unicode.
        convert_types (list, optional): Python types that should be converted
            to unicode. Defaults to the string types for the current Python
            version ([str, unicode, bytearray] in Python 2, [bytes, str,
            bytearray] otherwise).
        allow_pass (bool, optional): If True and the provided object is not
            converted, no error will be raised. Defaults to False and objects
            that cannot be converted will raise an error.

    Returns:
        str: Str version of input.

    Raises:
       TypeError: If supplied type cannot be converted to str.

    """
    if PY2:  # pragma: Python 2
        s_out = as_bytes(s_in, recurse=recurse, convert_types=convert_types,
                         allow_pass=allow_pass)
    else:  # pragma: Python 3
        s_out = as_unicode(s_in, recurse=recurse, convert_types=convert_types,
                           allow_pass=allow_pass)
    return s_out
    

def match_stype(s1, s2, **kwargs):
    r"""Encodes one string to match the type of the second.

    Args:
        s1 (str, bytes, bytearray, unicode): Object that type should be taken
            from.
        s2 (str, bytes, bytearray, unicode): Object that should be returned
            in the type from s1.
        **kwargs: Additional keyword arguments are passed to the conversion function.

    Returns:
        str, bytes, bytearray, unicode: Type matched version of s2.

    Raises:
        TypeError: If s1 is not str, bytes, bytearray or unicode.

    """
    if isinstance(s1, string_type):
        out = as_str(s2, **kwargs)
    elif isinstance(s1, bytes_type):
        out = as_bytes(s2, **kwargs)
    elif isinstance(s1, unicode_type):
        out = as_unicode(s2, **kwargs)
    elif isinstance(s1, bytearray):
        out = bytearray(as_unicode(s2), 'utf-8')
        if kwargs:  # pragma: debug
            raise RuntimeError("No conversion function for bytearray.")
    else:
        raise TypeError("Cannot match s1 type of '%s'" % type(s1))
    return out


def format_bytes(s, args):
    r"""Perform format on bytes/str, converting arguments to type of format
    string to ensure there is no prefix. For Python 3.4, if the format string
    is bytes, the formats and arguments will be changed to str type before
    format and then the resulting string will be changed back to bytes.

    Args:
        s (str, bytes): Format string.
        args (tuple): Arguments to be formated using the format string.

    Returns:
        str, bytes: Formatted argument string.

    """
    if PY2:  # pragma: Python 2
        out = s % args
    else:  # pragma: Python 3
        is_bytes = isinstance(s, bytes)
        new_args = []
        if PY34 or not is_bytes:
            converter = as_unicode
        else:
            converter = as_bytes
        for a in args:
            if isinstance(a, (bytes, str)):
                new_args.append(converter(a))
            else:
                new_args.append(a)
        out = converter(s) % tuple(new_args)
        if is_bytes:
            out = as_bytes(out)
    return out


def encode_escape(s):
    r"""Encode escape sequences.

    Args:
        s (str): String that should be encoded.

    Returns:
        str: Result of encoding escape sequences.

    """
    if PY2:  # pragma: Python 2
        out = match_stype(s, as_str(s).encode('string-escape'))
    else:  # pragma: Python 3
        out = match_stype(s, as_unicode(s).encode('unicode-escape'))
    return out


def decode_escape(s):
    r"""Decode escape sequences.

    Args:
        s (str): String that should be decoded.

    Returns:
        str: Result of decoding escape sequences.

    """
    if PY2:  # pragma: Python 2
        out = match_stype(s, as_bytes(s).decode('string-escape'))
    else:  # pragma: Python 3
        out = match_stype(s, as_bytes(s).decode('unicode-escape'))
    return out


# Python 3 version of np.genfromtxt
# https://github.com/numpy/numpy/issues/3184

    
__all__ = ['pickle', 'configparser', 'sio',
           'as_str', 'as_bytes', 'as_unicode']
